/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 *  Gabber
 *  Copyright (C) 1999-2001 Dave Smith & Julian Missig
 */

#ifndef INCL_GABBER_UTILITY_HH
#define INCL_GABBER_UTILITY_HH

#include "fwd.h"
#include <jabberoo/judo.hpp>
#include <jabberoo/presence.hh>
#include <jabberoo/roster.hh>

#include <glibmm/ustring.h>

namespace Gabber
{

/**
 * Utility Functions and types
 */
namespace Util
{
     void main_dialog(Gtk::Dialog* d);
     gint update_progress_bar(Gtk::ProgressBar* bar);
     void toggle_visible(Gtk::ToggleButton* t, Gtk::Widget* w);
//     gint strcasecmp_clist_items(GtkCList* c, const void* lhs, const void* rhs);
//     gint strcmp_clist_items(GtkCList* c, const void* lhs, const void* rhs);
     Glib::ustring substitute(const Glib::ustring& full_string, const Glib::ustring& var1);
     Glib::ustring substitute(const Glib::ustring& full_string, const Glib::ustring& var1, 
			      const Glib::ustring& var2);
     Glib::ustring substitute(const Glib::ustring& full_string, const Glib::ustring& var1, 
			      const Glib::ustring& var2, const Glib::ustring& var3);
//     void change_pixmap(Gtk::PixmapMenuItem* pitem, const string& filename);
     // We still have the widget parameter because we may move back to needing that at some point
     /**
      * Lookup a nickname given a JabberID and a Roster.
      * XXX: Maybe this should be in JabberOO?
      * @param r The Roster to use to find the nickname.
      * @param jid The JabberID to lookup.
      * @return Glib::ustring containing the nickname.
      */
     Glib::ustring lookup_nickname(const jabberoo::Roster& r, const Glib::ustring& jid);
     Glib::ustring getShowName(jabberoo::Presence::Show type);
     std::string   getShowFilename(jabberoo::Presence::Show type);
     Glib::ustring getS10nName(jabberoo::Roster::Subscription type);
     Glib::ustring getS10nInfo(jabberoo::Roster::Subscription type);
     jabberoo::Presence::Show indexShow(int show_index);
     int    indexShow(jabberoo::Presence::Show show_index);
     jabberoo::Presence::Type indexType(int type_index);
     int    indexType(jabberoo::Presence::Type type_index);
    std::string int2string(int i);
//     ConfigManager::Proxy indexProxy(int proxy_index);
//     int    indexProxy(ConfigManager::Proxy proxy_index);

    // Some common templates
    struct ListDeleter
    {
        template <typename T>
        void operator()(const T* ptr) { delete ptr; }
    };

    /** Predicate to test for a node in the specified namespace */
    struct MatchNamespace : public std::unary_function<judo::Node*, bool>
    {
        MatchNamespace(const std::string& ns_) : ns(ns_)
        { }
        bool operator()(const judo::Node* node)
        {
            return static_cast<const judo::Element*>(node)->cmpAttrib("xmlns", ns);
        }
        std::string ns;
    };

    typedef std::list<Glib::ustring> JIDList;
}; // namespace Util

//class GabberDnDText
//     : public SigC::Object
//{
//public:
//     GabberDnDText(Gtk::Text* widget);
//protected:
//     void on_drag_data_received(GdkDragContext* drag_ctx, gint x, gint y, GtkSelectionData* data, guint info, guint time);
//private:
//     Gtk::Text* _textWidget;
//};
//
//class MenuBuilder
//     : public SigC::Object
//{
//public:
//     MenuBuilder();
//     Gtk::Menu* get_menu() { return menu; }
//     void add_tearoff();
//     void add_separator();
//     Gtk::MenuItem* add_item(const string& label, int num, bool submenu = false);
//     Gtk::PixmapMenuItem* add_item(const string& pixmap, const string& label, int num, bool submenu = false);
//     void add_presence_items();
//     void finish_items();
//private:
//    Gtk::Menu* menu;
//public:
//     SigC::Signal1<void, int> selected;
//};
//
//class PresenceMenuBuilder
//     : public SigC::Object
//{
//public:
//     PresenceMenuBuilder();
//     Gtk::Menu* get_menu() { return _menu; }
//private:
//     MenuBuilder _main;
//     Gtk::Menu* _menu;
//     MenuBuilder _online;
//     MenuBuilder _chat;
//     MenuBuilder _away;
//     MenuBuilder _xa;
//     MenuBuilder _dnd;
//public:
//     SigC::Signal1<void, int> main_selected;
//     SigC::Signal2<void, int, jabberoo::Presence::Show> presence_selected;
//};
//
//class FileSel
//     : public SigC::Object
//{
//public:
//     FileSel::FileSel(const string& filename, const string& title, const string& log);
//protected:
//     void FileSel::on_ok_clicked();
//     void FileSel::on_cancel_clicked();
//private:
//     const string& _log;
//     Gtk::FileSelection* _fileSel;
//};
//

}; // namespace Gabber

#endif
