/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Gabber 2
 *  Based on Gabber, Copyright (c) 1999-2002 Dave Smith & Julian Missig
 *  Copyright (c) 2003 Julian Missig
 */

#include "StatusIcon.hh"

#include "GabberApp.hh"
#include "GabberWin.hh"
#include "JabberConnection.hh"
#include "ResourceManager.hh"

#include "intl.h"
#include <jabberoo/presence.hh>
#include <gtkmm/eventbox.h>
#include <gtkmm/image.h>

namespace Gabber {

StatusIcon::StatusIcon(GabberWin* gwin)
     : _gwin(gwin)
{
     // XXX: Find out if we have a status tray *at all*
     // if not, then every minute or so we should check again and then init
     init();
}

StatusIcon::~StatusIcon()
{
     if (_status_icon != NULL)
     {
	  delete _img_status;
	  delete _evt_queue;
          delete _status_icon;
	  g_free(_status_icon_egg);
     }
}

void StatusIcon::init()
{
     _status_icon_egg = egg_tray_icon_new(_("Gabber"));
     _status_icon = Glib::wrap(GTK_CONTAINER(_status_icon_egg));
//     _status_icon ->signal_delete_event().connect(slot(*this, &StatusIcon::on_delete_event));

     // Add the event box
     _evt_queue  = new Gtk::EventBox();
//     _evt_queue  ->set_shadow_type(Gtk::SHADOW_NONE);
     _status_icon->add(*_evt_queue);
     _evt_queue  ->add_events(Gdk::BUTTON_PRESS_MASK);
     _evt_queue  ->signal_button_press_event().connect(slot(*this, &StatusIcon::on_button_press));


     _img_status = new Gtk::Image(ResourceManager::getSingleton().getPixbuf("offline.png"));
     _evt_queue  ->add(*_img_status);
     _img_status ->show();
     _evt_queue  ->show();
     _status_icon->show();

     // Now that we've established all the widgets and whatnot
     // let's hook up to the events we want
     G_App.getSession().evtMyPresence.connect(slot(*this, &StatusIcon::on_my_presence));
     //_app.evtConnecting.connect(  SigC::slot(*this, &GabberWin::on_evtConnecting));
     //_app.evtConnected.connect(   SigC::slot(*this, &GabberWin::on_evtConnected));
     //_app.evtDisconnected.connect(SigC::slot(*this, &GabberWin::on_evtDisconnected));
}

void StatusIcon::on_my_presence(const jabberoo::Presence& p)
{
     // Set the labels
     std::string filename;
     if((p.getShow() == jabberoo::Presence::stInvalid) ||
        (p.getShow() == jabberoo::Presence::stOffline) ||
        p.getShow_str().empty())
          filename = "offline.png";
     else
          filename = p.getShow_str() + ".png";
     _img_status->set(ResourceManager::getSingleton().getPixbuf(filename));

     // Set the tooltip
     Glib::ustring tooltip;
     tooltip += Gabber::Util::getShowName(p.getShow()) + ": " + p.getStatus();
     _tips.unset_tip(*_evt_queue);
     _tips.set_tip(*_evt_queue, tooltip);
}

bool StatusIcon::on_button_press(GdkEventButton* e)
{
     // Determine if we should show/hide the main window or popup menu
     if (e->type == GDK_BUTTON_PRESS)
     {
          // XXX: if we have pending events, clicking triggers them
	  if (e->button == 1)
               _gwin->toggle_visibility();
          else if (e->button == 3)
          {
               // XXX: popup menu
          }
     }
     return false;
}

} // namespace Gabber
